/*--------------------------------------------------------------------
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>

#include "uvspec.h"
#include "../src_c/ascii.h"

/************************************/
/* prototypes of internal functions */
/************************************/

static int redistribute_moment_noalloc (float ****moment, int ***nmom, int nlambda, 
				int nlambda_lower, int nlambda_upper,
				float *z_old, int nlyr_old, 
				float *z_new, int nlyr_new);

int redistribute_molecular (atm_optprop_struct *optprop, int nlambda, 
			    int nlambda_lower, int nlambda_upper,
			    int *nq, 
			    float *zd, int nlyr, 
			    float *zd_common, int nlyr_common);

int redistribute_molecular_crs (crs_out_struct *crs, int nlambda, 
			    int nmol, 
			    float *zd, int nlyr, 
			    float *zd_common, int nlyr_common);

int redistribute_amf (atm_microphys_struct *microphys, 
		      crs_out_struct *crs,
		      int nlambda, 
		      int nlambda_lower, int nlambda_upper,
		      float *zd, int nlyr, 
		      float *zd_common, int nlyr_common);

/**************************************************************/
/* Setup redistribute irradiance.                         */
/**************************************************************/

int setup_redistribute (input_struct input, output_struct *output)
{
  int lc=0, ialt=0, i=0, status=0, old_nlev=0, ic=0;
  float *tmp_zd=NULL;


  int iv=0;
  double babso_o3=0, babso_co2=0, babso_o2=0, babso_no2=0;
  double babso_bro=0, babso_oclo=0, babso_hcho=0, babso_md=0;
  double babso_amf=0, babso_amftot=0, vertical_column=0, crs_amf=0;
  double babsotot_o3=0, babsotot_co2=0, babsotot_o2=0, babsotot_no2=0;
  double babsotot_bro=0, babsotot_oclo=0, babsotot_hcho=0;
  double vcol_o3=0, vcol_co2=0, vcol_o2=0, vcol_no2=0, vcol_bro=0;
  double vcol_oclo=0, vcol_hcho=0;

  double deltaz=0;

  aer_out_struct *redistributed_aer= calloc(1, sizeof(aer_out_struct));
  cld_out_struct *redistributed_ic = calloc(1, sizeof(cld_out_struct));
  cld_out_struct *redistributed_wc = calloc(1, sizeof(cld_out_struct));
  
  /************************ Redistribute molecular optical properties *******************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing molecular optical properties\n");
    fflush (stderr);
  }
  
  status = redistribute_molecular (&(output->atm.optprop), output->wl.nlambda_t, 
				   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				   output->atm.nq_t,
				   output->atm.zd, output->atm.nlev-1, 
				   output->atm.zd_common, output->atm.nlev_common-1);

  if (status!=0) {
    fprintf (stderr, "Error regridding molecular scattering and absorption\n");
    return status;
  }

  /************************ Redistribute molecular cross sections *******************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing molecular cross sections\n");
    fflush (stderr);
  }
  
  status = redistribute_molecular_crs (&(output->crs), output->wl.nlambda_t, 
				   MOL_NN,
				   output->atm.zd, output->atm.nlev-1, 
				   output->atm.zd_common, output->atm.nlev_common-1);

  if (status!=0) {
    fprintf (stderr, "Error regridding molecular cross sections\n");
    return status;
  }

  /******************** Redistribute aerosol stuff ****************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing aerosol optical properties\n");
    fflush (stderr);
  }

  calloc_aer_out(redistributed_aer, output->wl.nlambda_t, output->atm.nlev_common);
  status = cp_aer_out(&(output->aer), redistributed_aer, output->wl.nlambda_t, 1, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying output->aer to redistributed_aer\n", status);
    return status;
  }

  old_nlev = output->aer.nlev;
  tmp_zd = calloc (old_nlev+1, sizeof (float));
  for (lc=0;lc<old_nlev;lc++)  tmp_zd[lc] = output->aer.zd[lc]; 
  free_aer_out(&(output->aer), output->wl.nlambda_t);


  status = redistribute_optprop_noalloc (&(redistributed_aer->optprop), output->wl.nlambda_t, 
  				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
  				 tmp_zd, old_nlev-1, 
  				 output->atm.zd_common, output->atm.nlev_common-1);
  
  calloc_aer_out(&(output->aer), output->wl.nlambda_t, output->atm.nlev_common);
  status = cp_aer_out(redistributed_aer, &(output->aer), output->wl.nlambda_t, 1, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying redistributed_aer to output->aer\n", status);
    return status;
  }
  free_aer_out(redistributed_aer, output->wl.nlambda_t);
  free(tmp_zd);
  
  if (status!=0) {
    fprintf (stderr, "Error regridding aerosol parameters\n");
    return status;
  }

  /* copy number of levels and re-write altitude grid */
  output->aer.nlev = output->atm.nlev_common;
  free (output->aer.zd);
  output->aer.zd = calloc (output->aer.nlev, sizeof(float));
  for (lc=0; lc<output->aer.nlev; lc++)
    output->aer.zd[lc] = output->atm.zd_common[lc];

  /**************************** Redistribute water cloud stuff *************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing water cloud optical properties\n");
    fflush (stderr);
  }

    
  /********************************/
  /* loop over independent pixels */
  /********************************/
  if (input.wcipa) {
    for (ic=0; ic<output->nipa; ic++) {
      
      calloc_cld_out(redistributed_wc, output->wl.nlambda_t, output->atm.nlev_common);
      
      status = cp_cld_out(redistributed_wc, output->wcipa[ic], 0, 1, input.quiet);
      if (status!=0) {
	fprintf (stderr, "Error %d copying output->wcipa[%d] to redistributed_wc\n", ic, status);
	return status;
      }

      old_nlev = output->wcipa[ic].nlev;
      tmp_zd = calloc (old_nlev+1, sizeof (float));
      for (lc=0; lc<old_nlev; lc++)
	tmp_zd[lc] = output->wcipa[ic].zd[lc];
      free_cld_out (&(output->wcipa[ic]));

      status = redistribute_optprop_noalloc (&(redistributed_wc->optprop), output->wl.nlambda_t, 
					     output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
					     tmp_zd, old_nlev-1, 
					     output->atm.zd_common, output->atm.nlev_common-1);

      /* also redistribute cloud microphysics (required by MYSTIC) */
      if (output->wc.microphys.alloc) {
	status = redistribute_cld_microphys_noalloc (&(redistributed_wc->microphys),
						     tmp_zd, old_nlev-1, 
						     output->atm.zd_common, output->atm.nlev_common-1);
	
	if (status!=0) {
	  fprintf (stderr, "Error regridding water cloud microphysical parameters\n");
	  return status;
	}
      }

      calloc_cld_out(&(output->wcipa[ic]), output->wl.nlambda_t, output->atm.nlev_common);
      status = cp_cld_out(&(output->wcipa[ic]), *redistributed_wc, 0, 1, input.quiet);
      if (status!=0) {
	fprintf (stderr, "Error %d copying redistributed_wc to output->wcipa[%d]\n", ic, status);
	return status;
      }

      if (status!=0) {
	fprintf (stderr, "Error regridding water cloud parameters\n");
	return status;
      }

      free_cld_out(redistributed_wc);
      free(tmp_zd);

      /* copy number of levels and re-write altitude grid */
      output->wcipa[ic].nlev = output->atm.nlev_common;
      free (output->wcipa[ic].zd);
      output->wcipa[ic].zd = calloc (output->wcipa[ic].nlev, sizeof(float));
      for (lc=0; lc<output->wcipa[ic].nlev; lc++)
	output->wcipa[ic].zd[lc] = output->atm.zd_common[lc];
 
    }

    /* We also need a dummy output->wc at the appropriate resolution. Fudge it this way??? */
    /* NOOOOOO! We don't need a dummy but rather a REAL cloud because the optical          */
    /* properties are stored here, if explicit data are defined with wc_properties ...     */
    /*
    free_cld_out(&(output->wc));
    calloc_cld_out(&(output->wc), output->wl.nlambda_t, output->atm.nlev_common);
    status = cp_cld_out(&(output->wc), output->wcipa[0], 0, 1, input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d copying output->wcipa[0] to output->wc\n", status);
      return status;
    }
    */

    /* We need to redistribute output->wc in any case, hence we remove the following */
    /* "else" statement so that output->wc is also redistributed in case of ipa      */
  }
  /* else { */
  /* ??????????? ok ??????????? */
  calloc_cld_out(redistributed_wc, output->wl.nlambda_t, output->atm.nlev_common);

  status = cp_cld_out(redistributed_wc, output->wc, 1, 1, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying output->wc to redistributed_wc\n", status);
    return status;
  }

  old_nlev = output->wc.nlev;
  tmp_zd = calloc (old_nlev+1, sizeof (float));
  for (lc=0; lc<old_nlev; lc++) 
    tmp_zd[lc] = output->wc.zd[lc]; 
  free_cld_out(&(output->wc));
  
  status = redistribute_optprop_noalloc (&(redistributed_wc->optprop), output->wl.nlambda_t, 
					 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
					 tmp_zd, old_nlev-1, 
					 output->atm.zd_common, output->atm.nlev_common-1);
  
  /* also redistribute cloud microphysics (required by MYSTIC) */
  if (output->wc.microphys.alloc) {
    status = redistribute_cld_microphys_noalloc (&(redistributed_wc->microphys),
						 tmp_zd, old_nlev-1, 
						 output->atm.zd_common, output->atm.nlev_common-1);
    
    if (status!=0) {
      fprintf (stderr, "Error regridding water cloud microphysical parameters\n");
      return status;
    }
  }

  /* copy temporary cloud structure to final destination */
  calloc_cld_out(&(output->wc), output->wl.nlambda_t, output->atm.nlev_common);
  status = cp_cld_out(&(output->wc), *redistributed_wc, 1, 1, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying redistributed_wc to output->wc\n", status);
    return status;
  }
  
  free_cld_out(redistributed_wc);
  free(tmp_zd);

  /* } */

  /* copy number of levels and re-write altitude grid */
  output->wc.nlev = output->atm.nlev_common;
  free (output->wc.zd);
  output->wc.zd = calloc (output->wc.nlev, sizeof(float));
  for (lc=0; lc<output->wc.nlev; lc++)
    output->wc.zd[lc] = output->atm.zd_common[lc];
  

  /**************************** Redistribute ice cloud stuff ***************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing ice cloud optical properties\n");
    fflush (stderr);
  }

  if (input.icipa) {
    for (ic=0; ic<output->nipa; ic++) {
      
      calloc_cld_out(redistributed_ic, output->wl.nlambda_t, output->atm.nlev_common);
      
      status = cp_cld_out(redistributed_ic, output->icipa[ic], 0, 1, input.quiet);
      if (status!=0) {
	fprintf (stderr, "Error %d copying output->icipa[%d] to redistributed_ic\n", ic, status);
	return status;
      }
      old_nlev = output->icipa[ic].nlev;
      tmp_zd = calloc (old_nlev+1, sizeof (float));
      for (lc=0;lc<old_nlev;lc++) tmp_zd[lc] = output->icipa[ic].zd[lc];
      free_cld_out(&(output->icipa[ic]));

      status = redistribute_optprop_noalloc (&(redistributed_ic->optprop), output->wl.nlambda_t, 
					     output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
					     tmp_zd, old_nlev-1, 
					     output->atm.zd_common, output->atm.nlev_common-1);

      /* also redistribute cloud microphysics (required by MYSTIC) */
      if (output->ic.microphys.alloc) {
	status = redistribute_cld_microphys_noalloc (&(redistributed_ic->microphys),
						     tmp_zd, old_nlev-1, 
						     output->atm.zd_common, output->atm.nlev_common-1);
	
	if (status!=0) {
	  fprintf (stderr, "Error regridding ice cloud microphysical parameters\n");
	  return status;
	}
      }
      calloc_cld_out(&(output->icipa[ic]), output->wl.nlambda_t, output->atm.nlev_common);
      status = cp_cld_out(&(output->icipa[ic]), *redistributed_ic, 0, 1, input.quiet);
      if (status!=0) {
	fprintf (stderr, "Error %d copying redistributed_ic to output->icipa[%d]\n", ic, status);
	return status;
      }

      if (status!=0) {
	fprintf (stderr, "Error regridding ice cloud parameters\n");
	return status;
      }

      free_cld_out(redistributed_ic);
      free(tmp_zd);

      /* copy number of levels and re-write altitude grid */
      output->icipa[ic].nlev = output->atm.nlev_common;
      free (output->icipa[ic].zd);
      output->icipa[ic].zd = calloc (output->icipa[ic].nlev, sizeof(float));
      for (lc=0; lc<output->icipa[ic].nlev; lc++)
	output->icipa[ic].zd[lc] = output->atm.zd_common[lc];
 
    }

    /* We also need a dummy output->ic at the appropriate resolution. Fudge it this way??? */
    /* NOOOOOO! We don't need a dummy but rather a REAL cloud because the optical          */
    /* properties are stored here, if explicit data are defined with ic_properties ...     */
    /*
    free_cld_out(&(output->ic));
    calloc_cld_out(&(output->ic), output->wl.nlambda_t, output->atm.nlev_common);
    status = cp_cld_out(&(output->ic), output->icipa[0], 0, 1, input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d copying output->icipa[0] to output->ic\n", status);
      return status;
    }
    */

    /* We need to redistribute output->ic in any case, hence we remove the following */
    /* "else" statement so that output->ic is also redistributed in case of ipa      */
  }
  /* else { */
  /* ??????????? ok ??????????? */
  calloc_cld_out(redistributed_ic, output->wl.nlambda_t, output->atm.nlev_common);

  status = cp_cld_out(redistributed_ic, output->ic, 1, 1, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying output->ic to redistributed_ic\n", status);
    return status;
  }

  old_nlev = output->ic.nlev;
  tmp_zd = calloc (old_nlev+1, sizeof (float));
  for (lc=0;lc<old_nlev;lc++)  
    tmp_zd[lc] = output->ic.zd[lc]; 
  free_cld_out(&(output->ic));
  
  status = redistribute_optprop_noalloc (&(redistributed_ic->optprop), output->wl.nlambda_t, 
					 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
					 tmp_zd, old_nlev-1, 
					 output->atm.zd_common, output->atm.nlev_common-1);
  
  /* also redistribute cloud microphysics (required by MYSTIC) */
  if (output->ic.microphys.alloc) {
    status = redistribute_cld_microphys_noalloc (&(redistributed_ic->microphys),
						 tmp_zd, old_nlev-1, 
						 output->atm.zd_common, output->atm.nlev_common-1);
    
    if (status!=0) {
      fprintf (stderr, "Error regridding water cloud microphysical parameters\n");
      return status;
    }
  }


  /* copy temporary cloud structure to final destination */
  calloc_cld_out(&(output->ic), output->wl.nlambda_t, output->atm.nlev_common);
  status = cp_cld_out(&(output->ic), *redistributed_ic, 1, 1, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying redistributed_ic to output->ic\n", status);
    return status;
  }

  free_cld_out(redistributed_ic);
  free(tmp_zd);
  
  /* } */

  /* copy number of levels and re-write altitude grid */
  output->ic.nlev = output->atm.nlev_common;
  free (output->ic.zd);
  output->ic.zd = calloc (output->ic.nlev, sizeof(float));
  for (lc=0; lc<output->ic.nlev; lc++)
    output->ic.zd[lc] = output->atm.zd_common[lc];

  /************************* Interpolate pressure *************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating pressure\n");
    fflush (stderr);
  }

  status = interpolate_profile (output->atm.zd, &output->atm.microphys.press, output->atm.nlev,
				output->atm.zd_common, output->atm.nlev_common, 2);

  if (status!=0) {
    fprintf (stderr, "Error interpolating pressure\n");
    return status;
  }

  /************************* Interpolate densities ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating densities\n");
    fflush (stderr);
  }

  status=0;
  for (i=0; i<MOL_NN; i++) 
    status += interpolate_profile (output->atm.zd, &output->atm.microphys.dens[i], output->atm.nlev,
				   output->atm.zd_common, output->atm.nlev_common, 2);

  if (status!=0) {
    fprintf (stderr, "Error interpolating densities\n");
    return status;
  }

  /************************ Interpolate temperature ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating temperature\n");
    fflush (stderr);
  }

  status = interpolate_profile (output->atm.zd, &output->atm.microphys.temper, output->atm.nlev,
				output->atm.zd_common, output->atm.nlev_common, 1);

  if (status!=0) {
    fprintf (stderr, "Error interpolating temperature\n");
    return status;
  }

  /************************ Interpolate index of refraction ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  interpolating index of refraction\n");
    fflush (stderr);
  }

  status = interpolate_profile (output->atm.zd, &output->atm.microphys.refind, output->atm.nlev,
				output->atm.zd_common, output->atm.nlev_common, 1);

  if (status!=0) {
    fprintf (stderr, "Error regridding refractive index\n");
    return status;
  }

  /************************ Redistribute possible amf stuff ************************/

  if (input.verbose) {
    fprintf (stderr, " ...  redistributing airmass stuff\n");
    fflush (stderr);
  }

  status = redistribute_amf (&(output->atm.microphys), 
			     &(output->crs),
			     output->wl.nlambda_t, 
			     output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
			     output->atm.zd, output->atm.nlev-1, 
			     output->atm.zd_common, output->atm.nlev_common-1);

  if (status!=0) {
    fprintf (stderr, "Error regridding airmass stuff\n");
    return status;
  }

  /* Finally set new number of output level and layers, and set new zd */
  output->atm.nlev = output->atm.nlev_common;
  output->atm.nlyr = output->atm.nlev_common-1;
  free(output->atm.zd);
  output->atm.zd = calloc (output->atm.nlev, sizeof (float));
  for (lc=0; lc<output->atm.nlev_common; lc++) 
    output->atm.zd[lc] = output->atm.zd_common[lc];
  

  /* last but not least consider altitude */
  if (output->atm.altitude > 0) {
    for (ialt=0; ialt<output->atm.nlev_common; ialt++)
      if (output->atm.altitude == output->atm.zd_common[ialt])
	break;
    
    if (ialt == output->atm.nlev_common) {
      fprintf (stderr, "Fatal error, user-defined altitude %f not found in common grid\n", output->atm.altitude);
      return -1;
    }
	
    for (lc=0; lc<output->atm.nlev_common; lc++) {
      output->atm.zd_common[lc] -= output->atm.altitude;
      output->atm.zd[lc]        -= output->atm.altitude;
    }

    output->atm.nlev_common = ialt+1;
    output->atm.nlyr_common = ialt;
    output->atm.nlev        = ialt+1;
    output->atm.nlyr        = ialt;
  }


  /* something special here: we need to set the surface temperature  */
  /* to the atmospheric temperatzre at altitude; this has to be done */
  /* here rather than in setup_temperature                           */
  if (output->surface_temperature < 0) {
    output->surface_temperature = output->atm.microphys.temper[output->atm.nlev-1];
    
    if (input.verbose)
      fprintf (stderr, " ... setting surface temperature to %f K\n", output->surface_temperature);
  }
  
  

  /* Output vertical columns */
  if (input.verbose) {
    
    /* loop over wavelength */
    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      
      /* reset data */
      vcol_o3   = 0; 
      vcol_co2  = 0; 
      vcol_o2   = 0; 
      vcol_no2  = 0; 
      vcol_bro  = 0; 
      vcol_oclo = 0; 
      vcol_hcho = 0; 

      babsotot_o3   = 0; 
      babsotot_co2  = 0; 
      babsotot_o2   = 0; 
      babsotot_no2  = 0; 
      babsotot_bro  = 0; 
      babsotot_oclo = 0; 
      babsotot_hcho = 0; 
      babso_amftot  = 0;
      
      vertical_column=0;
      
      fprintf (stderr, "\n*** wavelength: iv = %d, %f nm\n", iv, output->wl.lambda_t[iv]);
      fprintf (stderr, "*** setup_redistribute()\n");
      fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------------\n");
      fprintf (stderr, "   lc |    z[km] |\n");
      fprintf (stderr, "      |          |           o3            co2             o2            no2            bro           oclo           hcho\n");
      fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------------\n");

      for (lc=0; lc<output->atm.nlyr; lc++) {
	
	/* The factor 1.e+5 converts z from km to cm.*/
	deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;
	
	babso_o3 = 0.5*deltaz * (output->atm.microphys.dens[MOL_O3][lc]*output->crs.crs[lc][MOL_O3][iv]
				 + output->atm.microphys.dens[MOL_O3][lc+1]*output->crs.crs[lc+1][MOL_O3][iv]);
	if ( output->atm.microphys.denstab_id != MOL_O3 ) babso_md += babso_o3;
	vcol_o3    += 0.5*deltaz * (output->atm.microphys.dens[MOL_O3][lc] + output->atm.microphys.dens[MOL_O3][lc+1]);
	

	babso_co2 = 0.5*deltaz * (output->atm.microphys.dens[MOL_CO2][lc]*output->crs.crs[lc][MOL_CO2][iv]
				 + output->atm.microphys.dens[MOL_CO2][lc+1]*output->crs.crs[lc+1][MOL_CO2][iv]);
	if ( output->atm.microphys.denstab_id != MOL_CO2 ) babso_md += babso_co2;
	vcol_co2    += 0.5*deltaz * (output->atm.microphys.dens[MOL_CO2][lc] + output->atm.microphys.dens[MOL_CO2][lc+1]);
	

	babso_o2 = 0.5*deltaz * (output->atm.microphys.dens[MOL_O2][lc]*output->crs.crs[lc][MOL_O2][iv]
				 + output->atm.microphys.dens[MOL_O2][lc+1]*output->crs.crs[lc+1][MOL_O2][iv]);
	if ( output->atm.microphys.denstab_id != MOL_O2 ) babso_md += babso_o2;
	vcol_o2    += 0.5*deltaz * (output->atm.microphys.dens[MOL_O2][lc] + output->atm.microphys.dens[MOL_O2][lc+1]);
	

	babso_no2 = 0.5*deltaz * (output->atm.microphys.dens[MOL_NO2][lc]*output->crs.crs[lc][MOL_NO2][iv]
				  + output->atm.microphys.dens[MOL_NO2][lc+1]*output->crs.crs[lc+1][MOL_NO2][iv]);
	if ( output->atm.microphys.denstab_id != MOL_NO2 ) babso_md += babso_no2;
	vcol_no2   += 0.5*deltaz * (output->atm.microphys.dens[MOL_NO2][lc] + output->atm.microphys.dens[MOL_NO2][lc+1]);
	

	babso_bro = 0.5*deltaz * (output->atm.microphys.dens[MOL_BRO][lc]*output->crs.crs[lc][MOL_BRO][iv]
				  + output->atm.microphys.dens[MOL_BRO][lc+1]*output->crs.crs[lc+1][MOL_BRO][iv]);
	if ( output->atm.microphys.denstab_id != MOL_BRO ) babso_md += babso_bro;
	vcol_bro   += 0.5*deltaz * (output->atm.microphys.dens[MOL_BRO][lc] + output->atm.microphys.dens[MOL_BRO][lc+1]);
	

	babso_oclo = 0.5*deltaz * (output->atm.microphys.dens[MOL_OCLO][lc]*output->crs.crs[lc][MOL_OCLO][iv]
				   + output->atm.microphys.dens[MOL_OCLO][lc+1]*output->crs.crs[lc+1][MOL_OCLO][iv]);
	if ( output->atm.microphys.denstab_id != MOL_OCLO ) babso_md += babso_oclo;
	vcol_oclo  += 0.5*deltaz * (output->atm.microphys.dens[MOL_OCLO][lc] + output->atm.microphys.dens[MOL_OCLO][lc+1]);
	

	babso_hcho = 0.5*deltaz * (output->atm.microphys.dens[MOL_HCHO][lc]*output->crs.crs[lc][MOL_HCHO][iv]
				   + output->atm.microphys.dens[MOL_HCHO][lc+1]*output->crs.crs[lc+1][MOL_HCHO][iv]);
	if ( output->atm.microphys.denstab_id != MOL_HCHO )  babso_md += babso_hcho;
	vcol_hcho  += 0.5*deltaz * (output->atm.microphys.dens[MOL_HCHO][lc] + output->atm.microphys.dens[MOL_HCHO][lc+1]);
	
	babsotot_o3   += babso_o3;
	babsotot_co2  += babso_co2;
	babsotot_o2   += babso_o2;
	babsotot_no2  += babso_no2;
	babsotot_bro  += babso_bro;
	babsotot_oclo += babso_oclo;
	babsotot_hcho += babso_hcho;
	babso_amftot  += babso_amf;
	
	fprintf (stderr, "%5d | %8.2f | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e\n", 
		 lc, output->atm.zd[lc+1],
		 babso_o3, babso_co2, babso_o2, babso_no2, babso_bro, babso_oclo, babso_hcho);
      }
      fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------------\n");
      fprintf (stderr, "%6s  %8.2f | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e\n", 
	       "AMFsum", 0.0/0.0,
	       babsotot_o3, babsotot_co2, babsotot_o2, babsotot_no2, babsotot_bro, babsotot_oclo, babsotot_hcho);
      fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------------\n");
      fprintf (stderr, "AMF column  %4.0f | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e\n", 
	       0.0/0.0, vcol_o3, vcol_co2, vcol_o2, vcol_no2, vcol_bro, vcol_oclo, vcol_hcho);
      fprintf (stderr, " ------------------------------------------------------------------------------------------------------------------------\n");
      fprintf (stderr, "AMF BOA crs %4.0f | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e | %12.6e\n", 
	       0.0/0.0,
	       output->crs.crs[output->atm.nlyr-1][MOL_O3]  [iv], 
	       output->crs.crs[output->atm.nlyr-1][MOL_CO2] [iv], 
	       output->crs.crs[output->atm.nlyr-1][MOL_O2]  [iv], 
	       output->crs.crs[output->atm.nlyr-1][MOL_NO2] [iv], 
	       output->crs.crs[output->atm.nlyr-1][MOL_BRO] [iv], 
	       output->crs.crs[output->atm.nlyr-1][MOL_OCLO][iv],
	       output->crs.crs[output->atm.nlyr-1][MOL_HCHO][iv]);
      fprintf (stderr, " -----------------------------------------------------------------------------------------------------------------------\n");
      if ( output->atm.microphys.denstab_id ) {
	fprintf(stderr,"Density matrix absorption column for airmass calculations: %16.10e %16.10e %16.10e\n",
		babso_amftot, vertical_column, crs_amf);
	fprintf (stderr, " -----------------------------------------------------------------------------------------------------------------------\n");
      }
    }    
  } 
  return status;
}

int compare_float (void *ap, void *bp) { /* Sort such that altitude is in descending order */
  float *a = (float *) ap;
  float *b = (float *) bp;
  
  if (*a == *b) 
    return 0;

  if (*a > *b) 
    return -1;
  else 
    return 1;
}

/****************************************************************/
/* Regrid data from an input grid to an output grid. All levels */
/* of the input grid must also be available in the output grid; */
/* the scale flag toggles between distributing a quantity over  */
/* the output levels (e.g. an optical thickness, scale==1)      */
/* or simply copying it (e.g. an asymmetry parameter, scale==0) */
/****************************************************************/

int regrid_float (float *xin, float  *yin,  int nin,
		  float *xout, float *yout, int nout,
		  int scale)
{
  int iin=0, iout=0;

  /* first check if both grids are sorted in descending order */
  for (iin=0; iin<nin-1; iin++)
    if (xin[iin] <= xin[iin+1]) {
      fprintf (stderr, "Error, input grid not sorted in descending order!\n");
      fprintf (stderr, "xin[%d] = %f, xin[%d] = %f\n", iin, xin[iin], iin+1, xin[iin+1]);
      return -1;
    }

  for (iout=0; iout<nout-1; iout++)
    if (xout[iout] <= xout[iout+1]) {
      fprintf (stderr, "Error, output grid not sorted in descending order!\n");
      fprintf (stderr, "xout[%d] = %f, xout[%d] = %f\n", iout, xout[iout], iout+1, xout[iout+1]);
      return -1;
    }

  /* first check if all grid points of the input grid */
  /* are also available in the output grid            */

  iout=nout-1;
  for (iin=nin-1; iin>=0; iin--) {
    while (xout[iout]<xin[iin]) 
      if (--iout<0)
	break;
      
    if (xin[iin]!=xout[iout]) {
      fprintf (stderr, "Error, level %f from input not contained in output\n",
	       xin[iin]);
      return -1;
    }
  }
  
  iin  = nin -1;
  iout = nout-1;

  /* go up to first output level */
  while (xout[iout]<xin[iin]) 
    iout--;

  for (; iout>=0; iout--) {

    while (xout[iout]<xin[iin-1]) {
      if (scale) 
	yout[iout] = yin[iin] * 
	  (xout[iout-1]-xout[iout]) / (xin[iin-1] - xin[iin]);
      else 
	yout[iout] = yin[iin];
      iout--;
    }

    iout++;
    iin--;
    
    if (iin<1)
      break;
  }

  return 0;
}


/*****************************************************************/
/* Regrid data from an input grid to an output grid. All levels  */
/* of the input grid must also be available in the output grid;  */
/* the scale flag toggles between distributing a quantity over   */
/* the output levels (e.g. an optical thickness, scale==1)       */
/* or simply copying it (e.g. an asymmetry parameter, scale==0); */
/*****************************************************************/

int regrid (double *xin, double  *yin,  int nin,
	    double *xout, double *yout, int nout,
	    int scale)
{
  int iin=0, iout=0;

  /* first check if both grids are sorted in descending order */
  for (iin=0; iin<nin-1; iin++)
    if (xin[iin] <= xin[iin+1]) {
      fprintf (stderr, "Error, input grid not sorted in descending order!\n");
      fprintf (stderr, "xin[%d] = %f, xin[%d] = %f\n", iin, xin[iin], iin+1, xin[iin+1]);
      return -1;
    }

  for (iout=0; iout<nout-1; iout++)
    if (xout[iout] <= xout[iout+1]) {
      fprintf (stderr, "Error, output grid not sorted in descending order!\n");
      fprintf (stderr, "xout[%d] = %f, xout[%d] = %f\n", iout, xout[iout], iout+1, xout[iout+1]);
      return -1;
    }

  /* first check if all grid points of the input grid */
  /* are also available in the output grid            */

  iout=nout-1;
  for (iin=nin-1; iin>=0; iin--) {
    while (xout[iout]<xin[iin]) 
      if (--iout<0)
	break;
      
    if (xin[iin]!=xout[iout]) {
      fprintf (stderr, "Error, level %16.12f from input not contained in output\n",
	       xin[iin]);
      return -1;
    }
  }

  
  iin  = nin -1;
  iout = nout-1;

  /* go up to first output level */
  while (xout[iout]<xin[iin]) 
    iout--;

  for (; iout>=0; iout--) {

    while (xout[iout]<xin[iin-1]) {
      if (scale) 
	yout[iout] = yin[iin] * 
	  (xout[iout-1]-xout[iout]) / (xin[iin-1] - xin[iin]);
      else 
	yout[iout] = yin[iin];
      iout--;
    }

    iout++;
    iin--;
    
    if (iin<1)
      break;
  }

  return 0;
}


/* copy array source to target and round to 6 digits */
void cp_round (float *source, float *target, int nlev)
{
  int lc=0;
  float pa=0;

  for (lc=0; lc<nlev; lc++) {
    if (source[lc]<=0.0)
      target[lc] = 0.0;
    else {
      pa=pow(10.0,floor(log10(source[lc]))-6.0);
      target[lc] = floor(source[lc]/pa)*pa;
    }
  }
}


void set_common_z_grid (float *zd, int nlev, float **zd_common, int *nlev_common) {
  int i=0, k=0, l=0, included=0;
  float *tmp = (float *) calloc (nlev+*nlev_common, sizeof(float));

  /* add the existing common grid to tmp */
  for (k=0; k<*nlev_common; k++)
    tmp[k] = (*zd_common)[k];

  /* cp_round (*zd_common, tmp, *nlev_common); */


  for (i=0; i<nlev; i++) {

    /* Is this level already included? */
    included = 0;
    for (l=0; l<k; l++)
      if (tmp[l] == zd[i]) {
	included=1;
	break;
      }
    
    /* if not already included then add */
    if (!included) 
      tmp[k++] = zd[i];
  }

  *nlev_common = k;

  if (*nlev_common==0) {   /* Special for first time, and no need to sort this */
    (*zd_common) = (float *) calloc (nlev, sizeof(float));

    for (i=0;i<nlev;i++) 
      (*zd_common)[i] = zd[i];      

    *nlev_common = nlev;
  }
  else {
    free(*zd_common);
    (*zd_common) = (float *) calloc (*nlev_common, sizeof(float));
    for (i=0; i<*nlev_common; i++) 
      (*zd_common)[i] = tmp[i];

    qsort((*zd_common), *nlev_common, sizeof(float), compare_float );
  }
  
  free(tmp);
}



/* add z levels from a file to an existing grid; */
/* reallocate memory as needed.                  */
int add_file2grid (float **zd, int *nlev, char *filename)
{
  int status=0;
  float *znew=NULL;
  int nnew=0;

  status = read_1c_file_float (filename, &znew, &nnew);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* add levels to aerosol profile */
  set_common_z_grid (znew, nnew, zd, nlev);
  free (znew);

  return 0;
}




/****************************************************************/
/* Regrid a 1D float array data from a given input grid (z_old) */
/* to an given output grid. All levels of the input grid must   */
/* also be available in the output grid; the scale flag toggles */
/* between distributing a quantity over the output levels       */
/* (e.g. an optical thickness, scale==1) or simply copying it   */
/* (e.g. an asymmetry parameter, scale==0); memory for data is  */
/* reallocated automatically.                                   */
/****************************************************************/

int redistribute_1D (void **data, 
		     float *z_old, int nlyr_old, 
		     float *z_new, int nlyr_new,
		     int scale, int type) 
{
  int lc=0, status=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double *tmp=NULL; 

  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+2, sizeof(double));
  yout = (double *) calloc (nlyr_new+2, sizeof(double));



  /* Allocate memory for temporary array */
  tmp = (double *) calloc (nlyr_old+1, sizeof(double));

  /* copy original stuff to temporary array */
  /* and reallocate with changed resolution */

  switch (type) {
  case REDISTRIBUTE_FLOAT:
    for (lc=0; lc<=nlyr_old; lc++)
      tmp[lc] = ((float *)  (*data))[lc];
    
    /* free original array and reallocate with changed resolution */
    free (*data);
    *data = calloc (nlyr_new, sizeof(float));

    break;

  case REDISTRIBUTE_DOUBLE:
    for (lc=0; lc<=nlyr_old; lc++)
      tmp[lc] = ((double *) (*data))[lc];

    free (*data);
    *data = calloc (nlyr_new, sizeof(double));

    break;
    
  default:
    fprintf (stderr, "Error, unknown data type %d in redistribute_1D()\n", type);
    return -1;
  }

  xin[0] = z_old[0];
  yin[0] = tmp[0];
  for (lc=0; lc<nlyr_old; lc++) {
    xin[lc+1]  = z_old[lc+1];
    yin[lc+1]  = tmp[lc];
  }

  for (lc=0; lc<=nlyr_new; lc++)
    xout[lc] = z_new[lc];
  
  status = regrid (xin, yin, nlyr_old+1,
		   xout, yout, nlyr_new+1, scale);
  
  if ( status != 0 ) {
    fprintf(stderr, "redistribute_1D(): Error %d returned by regrid()\n", status);
    return status;
  }
  
  switch (type) {
  case REDISTRIBUTE_FLOAT:
    for (lc=1; lc<=nlyr_new; lc++)
      ((float *) (*data))[lc-1] = yout[lc];      
    break;

  case REDISTRIBUTE_DOUBLE:
    for (lc=1; lc<=nlyr_new; lc++)
      ((double *) (*data))[lc-1] = yout[lc];      
    break;
    
  default:
    fprintf (stderr, "Error, unknown data type %d in redistribute_1D()\n", type);
    return -1;
  }

  free (tmp);
  
  free(xin); free(xout);
  free(yin); free(yout);
  
  return 0;
}

int redistribute_1D_noalloc (void **data, 
		     float *z_old, int nlyr_old, 
		     float *z_new, int nlyr_new,
		     int scale, int type) 
{
  int lc=0, status=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double *tmp=NULL; 

  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+2, sizeof(double));
  yout = (double *) calloc (nlyr_new+2, sizeof(double));


  /* Allocate memory for temporary array */
  tmp = (double *) calloc (nlyr_old+1, sizeof(double));

  /* copy original stuff to temporary array */

  switch (type) {
  case REDISTRIBUTE_FLOAT:
    for (lc=0; lc<=nlyr_old; lc++)
      tmp[lc] = ((float *)  (*data))[lc];
    
    break;

  case REDISTRIBUTE_DOUBLE:
    for (lc=0; lc<=nlyr_old; lc++)
      tmp[lc] = ((double *) (*data))[lc];

    break;
    
  default:
    fprintf (stderr, "Error, unknown data type %d in redistribute_1D()\n", type);
    return -1;
    }

  xin[0] = z_old[0];
  yin[0] = tmp[0];
  for (lc=0; lc<nlyr_old; lc++) {
    xin[lc+1]  = z_old[lc+1];
    yin[lc+1]  = tmp[lc];
  }

  for (lc=0; lc<=nlyr_new; lc++)
    xout[lc] = z_new[lc];
  
  status = regrid (xin, yin, nlyr_old+1,
		   xout, yout, nlyr_new+1, scale);
  
  if ( status != 0 ) {
    fprintf(stderr, "redistribute_1D(): Error %d returned by regrid()\n", status);
    return status;
  }
  
  switch (type) {
  case REDISTRIBUTE_FLOAT:
    for (lc=1; lc<=nlyr_new; lc++) 
      ((float *) (*data))[lc-1] = yout[lc];      
    break;

  case REDISTRIBUTE_DOUBLE:
    for (lc=1; lc<=nlyr_new; lc++)
      ((double *) (*data))[lc-1] = yout[lc];      
    break;
      
  default:
    fprintf (stderr, "Error, unknown data type %d in redistribute_1D()\n", type);
    return -1;
  }
  /*
  free (tmp);
  
  free(xin); free(xout);
  free(yin); free(yout);
  */
  return 0;
}



/****************************************************************/
/* Regrid a 2D float array data from a given input grid (z_old) */
/* to an given output grid. All levels of the input grid must   */
/* also be available in the output grid; the scale flag toggles */
/* between distributing a quantity over the output levels       */
/* (e.g. an optical thickness, scale==1) or simply copying it   */
/* (e.g. an asymmetry parameter, scale==0); memory for data is  */
/* reallocated automatically.                                   */
/****************************************************************/

int redistribute_2D (void ***data, int nlambda, 
		     int nlambda_lower, int nlambda_upper,
		     float *z_old, int nlyr_old, 
		     float *z_new, int nlyr_new,
		     int scale, int type) 
{
  int lc=0, iv=0, status=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double **tmp=NULL; 


  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+1, sizeof(double));
  yout = (double *) calloc (nlyr_new+1, sizeof(double));



  /* Allocate memory for temporary array */
  if ((status = ASCII_calloc_double (&tmp, nlambda, nlyr_old)) != 0) {
    fprintf (stderr, "Error allocating %d x %d doubles for tmp\n",
	     nlambda, nlyr_old);
    return status;
  }

  /* copy original stuff to temporary array */
  /* and reallocate with changed resolution */
  switch (type) {
  case REDISTRIBUTE_FLOAT:
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (lc=0; lc<nlyr_old; lc++) 
	tmp[iv][lc] = ((float **) (*data))[iv][lc];
    
    ASCII_free_float ((float **) (*data), nlambda);
    if ((status = ASCII_calloc_float ((float ***) data, nlambda, nlyr_new)) != 0) {
      fprintf (stderr, "Error allocating memory for data\n");
      return status;
    }
    break;

  case REDISTRIBUTE_DOUBLE:
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (lc=0; lc<nlyr_old; lc++) 
	tmp[iv][lc] = ((double **) (*data))[iv][lc];
    
    ASCII_free_double ((double **) (*data), nlambda);
    if ((status = ASCII_calloc_double ((double ***) data, nlambda, nlyr_new)) != 0) {
      fprintf (stderr, "Error allocating memory for data\n");
      return status;
    }
    break;
    
  default:
    fprintf (stderr, "Error, unknown data type %d in redistribute_2D()\n", type);
    return -1;
  }
  

  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    xin[0] = z_old[0];
    yin[0] = tmp[iv][0];
    for (lc=0; lc<nlyr_old; lc++) {
      xin[lc+1]  = z_old[lc+1];
      yin[lc+1]  = tmp[iv][lc];
    }
    
    for (lc=0; lc<=nlyr_new; lc++)
      xout[lc] = z_new[lc];
    
    status = regrid (xin, yin, nlyr_old+1,
		     xout, yout, nlyr_new+1, scale);

    if ( status != 0 ) {
      fprintf(stderr, "redistribute_2D(): Error %d returned by regrid()\n", status);
      return status;
    }

    switch (type) {
    case REDISTRIBUTE_FLOAT:
      for (lc=1; lc<=nlyr_new; lc++)
	(((float **) (*data))[iv][lc-1]) = yout[lc];      
      break;
      
    case REDISTRIBUTE_DOUBLE:
      for (lc=1; lc<=nlyr_new; lc++)
	(((double **) (*data))[iv][lc-1]) = yout[lc];      
      break;
      
    default:
      fprintf (stderr, "Error, unknown data type %d in redistribute_2D()\n", type);
      return -1;
    }
    
  }

  ASCII_free_double (tmp, nlambda);

  free(xin); free(xout);
  free(yin); free(yout);

  return 0;
}

int redistribute_2D_noalloc (void ***data, int nlambda, 
		     int nlambda_lower, int nlambda_upper,
		     float *z_old, int nlyr_old, 
		     float *z_new, int nlyr_new,
		     int scale, int type) 
{
  int lc=0, iv=0, status=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double **tmp=NULL; 

  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+1, sizeof(double));
  yout = (double *) calloc (nlyr_new+1, sizeof(double));



  /* Allocate memory for temporary array */
  if ((status = ASCII_calloc_double (&tmp, nlambda, nlyr_old)) != 0) {
    fprintf (stderr, "Error allocating %d x %d doubles for tmp\n",
	     nlambda, nlyr_old);
    return status;
  }

  /* copy original stuff to temporary array */
  switch (type) {
  case REDISTRIBUTE_FLOAT:
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (lc=0; lc<nlyr_old; lc++) 
	tmp[iv][lc] = ((float **) (*data))[iv][lc];
    
    break;

  case REDISTRIBUTE_DOUBLE:
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (lc=0; lc<nlyr_old; lc++) 
	tmp[iv][lc] = ((double **) (*data))[iv][lc];
    
    break;
    
  default:
    fprintf (stderr, "Error, unknown data type %d in redistribute_2D_noalloc()\n", type);
    return -1;
  }
  
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    xin[0] = z_old[0];
    yin[0] = tmp[iv][0];
    for (lc=0; lc<nlyr_old; lc++) {
      xin[lc+1]  = z_old[lc+1];
      yin[lc+1]  = tmp[iv][lc];
    }
    
    for (lc=0; lc<=nlyr_new; lc++)
      xout[lc] = z_new[lc];
    
    status = regrid (xin, yin, nlyr_old+1,
		     xout, yout, nlyr_new+1, scale);

    if ( status != 0 ) {
      fprintf(stderr, "redistribute_2D_noalloc(): Error %d returned by regrid()\n", status);
      return status;
    }

    switch (type) {
    case REDISTRIBUTE_FLOAT:
      for (lc=1; lc<=nlyr_new; lc++)
	(((float **) (*data))[iv][lc-1]) = yout[lc];      
      break;
      
    case REDISTRIBUTE_DOUBLE:
      for (lc=1; lc<=nlyr_new; lc++)
	(((double **) (*data))[iv][lc-1]) = yout[lc];      
      break;
      
    default:
      fprintf (stderr, "Error, unknown data type %d in redistribute_2D_noalloc()\n", type);
      return -1;
    }
    
  }

  ASCII_free_double (tmp, nlambda);

  free(xin); free(xout);
  free(yin); free(yout);

  return 0;
}



/*****************************************************************/
/* Regrid phase function moments from a given input grid (z_old) */
/* to an given output grid. All levels of the input grid must    */
/* also be available in the output grid; the scale flag toggles  */
/* between distributing a quantity over the output levels        */
/* (e.g. an optical thickness, scale==1) or simply copying it    */
/* (e.g. an asymmetry parameter, scale==0); memory for the       */
/* moments is reallocated automatically.                         */
/*****************************************************************/

static int redistribute_moment (float ****moment, int ***nmom, int nlambda, 
				int nlambda_lower, int nlambda_upper,
				float *z_old, int nlyr_old, 
				float *z_new, int nlyr_new) 
{
  int lc=0, iv=0, k=0, status=0;
  int maxmom=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double **tmp_mom=NULL, **tmp_newmom=NULL;
  int *tmp_nmom = calloc (nlyr_old, sizeof(int)); 


  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+1, sizeof(double));
  yout = (double *) calloc (nlyr_new+1, sizeof(double));


  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    
    /* copy original stuff to temporary array */
    for (lc=0; lc<nlyr_old; lc++) 
      tmp_nmom[lc] = (*nmom)[iv][lc];
    
    /* free original array and reallocate with changed resolution */
    free ((*nmom)[iv]);
    (*nmom)[iv] = calloc (nlyr_new, sizeof(int));
    

    /* regrid number of moments */
    xin[0] = z_old[0];
    yin[0] = (float) tmp_nmom[0];
    for (lc=0; lc<nlyr_old; lc++) {
      xin[lc+1]  = z_old[lc+1];
      yin[lc+1]  = tmp_nmom[lc];
    }
    
    for (lc=0; lc<=nlyr_new; lc++)
      xout[lc] = z_new[lc];

    status = regrid (xin, yin, nlyr_old+1,
		     xout, yout, nlyr_new+1, 0);
    
    if ( status != 0 ) {
      fprintf(stderr, "redistribute_moment(): Error %d returned by regrid()\n", status);
      return status;
    }
    
    for (lc=1; lc<=nlyr_new; lc++)
      (*nmom)[iv][lc-1] = (int) (yout[lc] + 0.5);

    
    /* determine maximum number of moments */
    maxmom = 0;
    for (lc=0; lc<nlyr_old; lc++) 
      if (tmp_nmom[lc] > maxmom) 
	maxmom = tmp_nmom[lc];
    
    /* Allocate memory for temporary array of moments */
    tmp_mom = calloc (nlyr_old, sizeof (double *));
    for (lc=0; lc<nlyr_old; lc++) 
      tmp_mom[lc] = calloc (maxmom+1, sizeof (double));
    
    /* copy original stuff to temporary array */
    for (lc=0; lc<nlyr_old; lc++) 
      for (k=0; k<tmp_nmom[lc]; k++) 
	tmp_mom[lc][k] = (*moment)[iv][lc][k];
    
    /* allocate temporary array for regridded moments */
    tmp_newmom = calloc (nlyr_new, sizeof (double *));
    for (lc=0; lc<nlyr_new; lc++) 
      tmp_newmom[lc] = calloc (maxmom+1, sizeof (double));
    
    /* free original moments array and allocate with changed resolution */
    ASCII_free_float ((*moment)[iv], nlyr_old);
    (*moment)[iv] = calloc (nlyr_new, sizeof (float *));
    for (lc=0; lc<nlyr_new; lc++) 
      (*moment)[iv][lc] = calloc ((*nmom)[iv][lc]+1, sizeof (float));
  
    for (k=0; k<maxmom; k++) {
      xin[0] = z_old[0];
      yin[0] = tmp_mom[0][k];
      for (lc=0; lc<nlyr_old; lc++) {
	xin[lc+1]  = z_old[lc+1];
	yin[lc+1]  = tmp_mom[lc][k];
      }
      for (lc=0; lc<=nlyr_new; lc++) 
	xout[lc] = z_new[lc];
      
      status = regrid (xin, yin, nlyr_old+1,
		       xout, yout, nlyr_new+1, 0);
      if ( status != 0 ) {
	fprintf(stderr, "Error %d returned by regrid()\n", status);
	return status;
      }
      for (lc=1; lc<=nlyr_new; lc++) 
	tmp_newmom[lc-1][k] = yout[lc];
    }
    
    for (lc=0; lc<nlyr_new; lc++) 
      for (k=0; k<(*nmom)[iv][lc]; k++) 
	(*moment)[iv][lc][k] = (float) tmp_newmom[lc][k];
    
    ASCII_free_double (tmp_mom,    nlyr_old);
    ASCII_free_double (tmp_newmom, nlyr_new);
  }
  
  free(tmp_nmom);
  
  free(xin); free(xout);
  free(yin); free(yout);

  return 0;
}


static int redistribute_moment_noalloc (float ****moment, int ***nmom, int nlambda, 
				int nlambda_lower, int nlambda_upper,
				float *z_old, int nlyr_old, 
				float *z_new, int nlyr_new) 
{
  int lc=0, iv=0, k=0, status=0;
  int maxmom=0;
  double *xin =NULL, *yin =NULL;
  double *xout=NULL, *yout=NULL;

  double **tmp_mom=NULL, **tmp_newmom=NULL;
  int *tmp_nmom = calloc (nlyr_old, sizeof(int)); 


  xin  = (double *) calloc (nlyr_old+1, sizeof(double));
  yin  = (double *) calloc (nlyr_old+1, sizeof(double));

  xout = (double *) calloc (nlyr_new+1, sizeof(double));
  yout = (double *) calloc (nlyr_new+1, sizeof(double));


  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    
    /* copy original stuff to temporary array */
    for (lc=0; lc<nlyr_old; lc++) 
      tmp_nmom[lc] = (*nmom)[iv][lc];
    

    /* regrid number of moments */
    xin[0] = z_old[0];
    yin[0] = (float) tmp_nmom[0];
    for (lc=0; lc<nlyr_old; lc++) {
      xin[lc+1]  = z_old[lc+1];
      yin[lc+1]  = tmp_nmom[lc];
    }
    
    for (lc=0; lc<=nlyr_new; lc++)
      xout[lc] = z_new[lc];

    status = regrid (xin, yin, nlyr_old+1,
		     xout, yout, nlyr_new+1, 0);
    
    if ( status != 0 ) {
      fprintf(stderr, "redistribute_moment_noalloc(): Error %d returned by regrid()\n", status);
      return status;
    }
    
    for (lc=1; lc<=nlyr_new; lc++)
      (*nmom)[iv][lc-1] = (int) (yout[lc] + 0.5);

    
    /* determine maximum number of moments */
    maxmom = 0;
    for (lc=0; lc<nlyr_old; lc++) 
      if (tmp_nmom[lc] > maxmom) 
	maxmom = tmp_nmom[lc];
    
    /* Allocate memory for temporary array of moments */
    tmp_mom = calloc (nlyr_old, sizeof (double *));
    for (lc=0; lc<nlyr_old; lc++) 
      tmp_mom[lc] = calloc (maxmom+1, sizeof (double));
    
    /* copy original stuff to temporary array */
    for (lc=0; lc<nlyr_old; lc++) 
      for (k=0; k<tmp_nmom[lc]; k++) 
	tmp_mom[lc][k] = (*moment)[iv][lc][k];
    
    /* allocate temporary array for regridded moments */
    tmp_newmom = calloc (nlyr_new, sizeof (double *));
    for (lc=0; lc<nlyr_new; lc++) 
      tmp_newmom[lc] = calloc (maxmom+1, sizeof (double));
      
    /* free original moments array and allocate with changed resolution */
    ASCII_free_float ((*moment)[iv], nlyr_old);
    (*moment)[iv] = calloc (nlyr_new, sizeof (float *));
    for (lc=0; lc<nlyr_new; lc++) 
      (*moment)[iv][lc] = calloc ((*nmom)[iv][lc]+1, sizeof (float));

    for (k=0; k<maxmom; k++) {
      xin[0] = z_old[0];
      yin[0] = tmp_mom[0][k];
      for (lc=0; lc<nlyr_old; lc++) {
	xin[lc+1]  = z_old[lc+1];
	yin[lc+1]  = tmp_mom[lc][k];
      }
      for (lc=0; lc<=nlyr_new; lc++) 
	xout[lc] = z_new[lc];
      
      status = regrid (xin, yin, nlyr_old+1,
		       xout, yout, nlyr_new+1, 0);
      if ( status != 0 ) {
	fprintf(stderr, "Error %d returned by regrid()\n", status);
	return status;
      }
      for (lc=1; lc<=nlyr_new; lc++) 
	tmp_newmom[lc-1][k] = yout[lc];
    }
    
    for (lc=0; lc<nlyr_new; lc++) 
      for (k=0; k<(*nmom)[iv][lc]; k++) 
	(*moment)[iv][lc][k] = (float) tmp_newmom[lc][k];
      
    
    ASCII_free_double (tmp_mom,    nlyr_old);
    ASCII_free_double (tmp_newmom, nlyr_new);
  }
  
  free(tmp_nmom);
  
  free(xin); free(xout);
  free(yin); free(yout);

  return 0;
}





/***********************************/
/* Redistribute optical properties */
/***********************************/

int redistribute_optprop (optprop_struct *optprop, int nlambda, 
			  int nlambda_lower, int nlambda_upper,
			  float *zd, int nlyr, 
			  float *zd_common, int nlyr_common)
{
  int status=0;

  /** dtau **/
  status += redistribute_2D ((void ***) (&(optprop->dtau)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     1, REDISTRIBUTE_FLOAT);

  /** ssa  **/
  status += redistribute_2D ((void ***) (&(optprop->ssa)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);

  /**  g1  **/
  status += redistribute_2D ((void ***) (&(optprop->g1)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);

  /**  g2  **/
  status += redistribute_2D ((void ***) (&(optprop->g2)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);

  /**  ff  **/
  status += redistribute_2D ((void ***) (&(optprop->ff)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);

  /** moments **/
  status += redistribute_moment (&(optprop->moment), &(optprop->nmom), nlambda, 
				 nlambda_lower, nlambda_upper,
				 zd, nlyr, 
				 zd_common, nlyr_common);
  
  return status;
}

int redistribute_optprop_noalloc (optprop_struct *optprop, int nlambda, 
			  int nlambda_lower, int nlambda_upper,
			  float *zd, int nlyr, 
			  float *zd_common, int nlyr_common)
{
  int status=0;

  /** dtau **/
  status += redistribute_2D_noalloc ((void ***) (&(optprop->dtau)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     1, REDISTRIBUTE_FLOAT);
  if (status!=0) {
    fprintf (stderr, "redistribute_optprop_noalloc: Error %d  redistributing dtau\n", status);
    return status;
  }

  /** ssa  **/
  status += redistribute_2D_noalloc ((void ***) (&(optprop->ssa)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);
  if (status!=0) {
    fprintf (stderr, "redistribute_optprop_noalloc: Error %d  redistributing ssa\n", status);
    return status;
  }

  /**  g1  **/
  status += redistribute_2D_noalloc ((void ***) (&(optprop->g1)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);
  if (status!=0) {
    fprintf (stderr, "redistribute_optprop_noalloc: Error %d  redistributing g1\n", status);
    return status;
  }

  /**  g2  **/
  status += redistribute_2D_noalloc ((void ***) (&(optprop->g2)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);
  if (status!=0) {
    fprintf (stderr, "redistribute_optprop_noalloc: Error %d  redistributing g2\n", status);
    return status;
  }

  /**  ff  **/
  status += redistribute_2D_noalloc ((void ***) (&(optprop->ff)), nlambda, 
			     nlambda_lower, nlambda_upper,
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);
  if (status!=0) {
    fprintf (stderr, "redistribute_optprop_noalloc: Error %d  redistributing ff\n", status);
    return status;
  }

  /** moments **/
  status += redistribute_moment_noalloc (&(optprop->moment), &(optprop->nmom), nlambda, 
					 nlambda_lower, nlambda_upper,
					 zd, nlyr, 
					 zd_common, nlyr_common);
  if (status!=0) {
    fprintf (stderr, "redistribute_optprop_noalloc: Error %d  redistributing moment\n", status);
    return status;
  }
  
  return status;
}


/***********************************/
/* Redistribute optical properties */
/***********************************/

int redistribute_cld_microphys_noalloc (cld_microphys_struct *microphys,
				float *zd, int nlyr, 
				float *zd_common, int nlyr_common)
{
  int status=0;
  
  /** lwc for layers  **/
  status += redistribute_1D_noalloc ((void **) (&(microphys->lwc_layer)),
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);

  /** reff for layers **/
  status += redistribute_1D_noalloc ((void **) (&(microphys->effr_layer)),
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);


  /** lwc at levels  **/
  status += redistribute_1D_noalloc ((void **) (&(microphys->lwc)),
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);
  
  /** reff at levels **/
  status += redistribute_1D_noalloc ((void **) (&(microphys->effr)),
			     zd, nlyr, 
			     zd_common, nlyr_common,
			     0, REDISTRIBUTE_FLOAT);


  /* ????? the regridded lwc and reff are not necessarily correct at the edges ????? */
  /* ????? and should therefore not be used after regridding; a special        ????? */
  /* ????? treatment is required to "conserve" the hard boundaries;            ????? */
  /* ????? use lwc_layer and effr_layer instead!                               ????? */ 

  return status;
}


/**************************************************/
/* Redistribute molecular cross sections          */
/**************************************************/


int redistribute_molecular_crs (crs_out_struct *crs, int nlambda, 
			    int nmol, 
			    float *zd, int nlyr, 
			    float *zd_common, int nlyr_common)
{
  int lc=0, iv=0, id=0, scale=0, status=0;
  double ***tmpcrs=NULL;

  double *xin=NULL, *yin=NULL, *xout=NULL, *yout=NULL;

  xin  = (double *) calloc (nlyr_common+1, sizeof(double));
  yin  = (double *) calloc (nlyr_common+1, sizeof(double));

  xout = (double *) calloc (nlyr_common+1, sizeof(double));
  yout = (double *) calloc (nlyr_common+1, sizeof(double));

  /* Allocate memory for temporary ary */
  if ((status = ASCII_calloc_double_3D (&tmpcrs, nlyr+1, nmol, nlambda)) != 0) {
    fprintf (stderr, "redistribute_molecular_crs: Error allocating memory for tmpcrs\n");
    return status;
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr; lc++) 
    for (id=0; id<nmol; id++)
      for (iv=0; iv<nlambda; iv++) {
	tmpcrs[lc][id][iv] = crs->crs[lc][id][iv];
      }
  
  /* free original ary and allocate with changed resolution */
  ASCII_free_float_3D (crs->crs, nlyr+1, nmol);
  if ((status = ASCII_calloc_float_3D(&crs->crs, nlyr_common+1, nmol, nlambda)) != 0) {
    fprintf (stderr, "redistribute_molecular_crs: Error allocating memory for crs->crs\n");
    return status;
  }

  scale = 0;
  for (iv=0; iv<nlambda; iv++) {
    for (id=0; id<nmol; id++) {
      xin[0] = zd[0];
      yin[0] = tmpcrs[0][id][iv];
      for (lc=0; lc<nlyr; lc++) {
	xin[lc+1] = zd[lc+1];
	yin[lc+1] = tmpcrs[lc][id][iv];
      }
      
      for (lc=0; lc<=nlyr_common; lc++) 
      	xout[lc] = zd_common[lc];

      status = regrid (xin, yin, nlyr+1,
		       xout, yout, nlyr_common+1, scale);
      if ( status != 0 ) {
	fprintf(stderr, "redistribute_molecular_crs: error calling regrid for crs\n");
	return status;
      }
      for (lc=1; lc<=nlyr_common; lc++) 
	crs->crs[lc-1][id][iv] = yout[lc];      
    }
  }
  ASCII_free_double_3D (tmpcrs, nlyr, nmol);

  /* free memory */
  free (xin); free (xout);
  free (yin); free (yout);

  return status;
}



/**************************************************/
/* Redistribute molecular scattering / absorption */
/**************************************************/

int redistribute_molecular (atm_optprop_struct *optprop, int nlambda, 
			    int nlambda_lower, int nlambda_upper,
			    int *nq, 
			    float *zd, int nlyr, 
			    float *zd_common, int nlyr_common)
{
  int lc=0, iv=0, iq=0, scale=0, status=0;
  double ***tmpopt=NULL;

  double *xin=NULL, *yin=NULL, *xout=NULL, *yout=NULL;

  xin  = (double *) calloc (nlyr_common+1, sizeof(double));
  yin  = (double *) calloc (nlyr_common+1, sizeof(double));

  xout = (double *) calloc (nlyr_common+1, sizeof(double));
  yout = (double *) calloc (nlyr_common+1, sizeof(double));


  /* Allocate memory for temporary ary */
  if ((status = ASCII_calloc_double_3D_arylen_restricted (&tmpopt, nlyr, 
							  nlambda, nlambda_lower, nlambda_upper, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (iq=0; iq<nq[iv]; iq++) 
	tmpopt[lc][iv][iq] = optprop->tau_molabs_t[lc][iv][iq];
  
  /* free original ary and allocate with changed resolution */
  ASCII_free_double_3D (optprop->tau_molabs_t, nlyr, nlambda);
  if ((status = ASCII_calloc_double_3D_arylen_restricted (&optprop->tau_molabs_t, nlyr_common, 
							  nlambda, nlambda_lower, nlambda_upper, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for optprop->tau_molabs_t\n");
    return status;
  }

  scale = 1;
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    for (iq=0; iq<nq[iv]; iq++) {
      xin[0] = zd[0];
      yin[0] = tmpopt[0][iv][iq];
      for (lc=0; lc<nlyr; lc++) {
	xin[lc+1] = zd[lc+1];
	yin[lc+1] = tmpopt[lc][iv][iq];
      }

      for (lc=0; lc<=nlyr_common; lc++) 
      	xout[lc] = zd_common[lc];

	status = regrid (xin, yin, nlyr+1,
			 xout, yout, nlyr_common+1, scale);
	if ( status != 0 ) {
	  fprintf(stderr, "Redistribute: error calling regrid for tau_molabs_t\n");
	  return status;
	}
	for (lc=1; lc<=nlyr_common; lc++) 
	  optprop->tau_molabs_t[lc-1][iv][iq] = yout[lc];      
    }
  }
  ASCII_free_double_3D (tmpopt, nlyr, nlambda);
  
  /************************ Redistribute molecular optical properties -md *********************/

  /* Allocate memory for temporary ary */
  if ((status = ASCII_calloc_double_3D_arylen_restricted (&tmpopt, nlyr, 
							  nlambda, nlambda_lower, nlambda_upper, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (iq=0; iq<nq[iv]; iq++) 
	tmpopt[lc][iv][iq] = optprop->tau_molabs_md_t[lc][iv][iq];
  
  /* free original ary and allocate with changed resolution */
  ASCII_free_double_3D (optprop->tau_molabs_md_t, nlyr, nlambda);
  if ((status = ASCII_calloc_double_3D_arylen_restricted (&optprop->tau_molabs_md_t, nlyr_common, 
							  nlambda, nlambda_lower, nlambda_upper, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for optprop->tau_molabs_md_t\n");
    return status;
  }

  scale = 1;
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    for (iq=0; iq<nq[iv]; iq++) {
      xin[0] = zd[0];
      yin[0] = tmpopt[0][iv][iq];
      for (lc=0; lc<nlyr; lc++) {
	xin[lc+1]  = zd[lc+1];
	yin[lc+1]  = tmpopt[lc][iv][iq];
      }
      for (lc=0; lc<=nlyr_common; lc++) 
      	xout[lc] = zd_common[lc];

	status = regrid (xin, yin, nlyr+1,
			 xout, yout, nlyr_common+1, scale);
	if ( status != 0 ) {
	  fprintf(stderr, "Redistribute: error calling regrid for tau_molabs_md\n");
	  return status;
	}
	for (lc=1; lc<=nlyr_common; lc++) 
	  optprop->tau_molabs_md_t[lc-1][iv][iq] = yout[lc];      
    }
  }
  ASCII_free_double_3D (tmpopt, nlyr, nlambda);

  /********************************** Redistribute Rayleigh scattering *********************/

  /* Allocate memory for temporary ary */
  if ((status = ASCII_calloc_double_3D_arylen_restricted (&tmpopt, nlyr, 
							  nlambda, nlambda_lower, nlambda_upper, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  /* cp original stuff to tmp ary */
  for (lc=0; lc<nlyr; lc++) 
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++)
      for (iq=0; iq<nq[iv]; iq++) 
	tmpopt[lc][iv][iq] = optprop->tau_rayleigh_t[lc][iv][iq];
  
  /* free original ary and allocate with changed resolution */
  ASCII_free_double_3D (optprop->tau_rayleigh_t, nlyr, nlambda);
  if ((status = ASCII_calloc_double_3D_arylen_restricted (&optprop->tau_rayleigh_t, nlyr_common, 
							  nlambda, nlambda_lower, nlambda_upper, nq)) != 0) {
    fprintf (stderr, "Error allocating memory for tmpopt\n");
    return status;
  }

  scale = 1;
  for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
    for (iq=0; iq<nq[iv]; iq++) {
      xin[0] = zd[0];
      yin[0] = tmpopt[0][iv][iq];
      for (lc=0; lc<nlyr; lc++) {
	xin[lc+1]  = zd[lc+1];
	yin[lc+1]  = tmpopt[lc][iv][iq];
      }
      for (lc=0; lc<=nlyr_common; lc++) 
      	xout[lc] = zd_common[lc];

	status = regrid (xin, yin, nlyr+1,
			 xout, yout, nlyr_common+1, scale);
	if ( status != 0 ) {
	  fprintf(stderr, "Redistribute: error calling regrid for tau_rayleigh_t\n");
	  return status;
	}
	for (lc=1; lc<=nlyr_common; lc++) 
	  optprop->tau_rayleigh_t[lc-1][iv][iq] = yout[lc];      
    }
  }
  ASCII_free_double_3D (tmpopt, nlyr, nlambda);

  /* free memory */
  free (xin); free (xout);
  free (yin); free (yout);

  return 0;
}



/******************************/
/* Redistribute airmass stuff */
/******************************/

int redistribute_amf (atm_microphys_struct *microphys, 
		      crs_out_struct *crs,
		      int nlambda, 
		      int nlambda_lower, int nlambda_upper,
		      float *zd, int nlyr, 
		      float *zd_common, int nlyr_common)
{
  int is=0, iv=0, lc=0, scale=0, status=0;
  

  float  *xfin=NULL, *yfin=NULL, *xfout=NULL, *yfout=NULL;
  double *xin=NULL, *yin=NULL, *xout=NULL, *yout=NULL;
  double **tmp2=NULL;


  xin  = (double *) calloc (nlyr_common+1, sizeof(double));
  yin  = (double *) calloc (nlyr_common+1, sizeof(double));
  xout = (double *) calloc (nlyr_common+1, sizeof(double));
  yout = (double *) calloc (nlyr_common+1, sizeof(double));

  xfin  = (float *) calloc (nlyr_common+1, sizeof(float));
  yfin  = (float *) calloc (nlyr_common+1, sizeof(float));
  xfout = (float *) calloc (nlyr_common+1, sizeof(float));
  yfout = (float *) calloc (nlyr_common+1, sizeof(float));



  if (microphys->denstab_id > 0) {

    /** crs_amf **/

    /* Allocate memory for temporary ary */
    if ((status = ASCII_calloc_double (&tmp2, nlambda, nlyr+1)) != 0) {
      fprintf (stderr, "Error allocating memory for tmp2\n");
      return status;
    }

    /* cp original stuff to tmp ary */
    for (lc=0; lc<nlyr+1; lc++) 
      for (iv=nlambda_lower; iv<=nlambda_upper; iv++) 
	tmp2[iv][lc] = crs->crs_amf[iv][lc];

    /* allocate needed memory */
    ASCII_free_float (crs->crs_amf, nlambda);
    if ((status = ASCII_calloc_float (&crs->crs_amf, nlambda,
				      nlyr_common+1)) != 0)
      return status;

    scale = 0;
    for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
      for (lc=0; lc<nlyr+1; lc++) {
	xin[lc]  = zd[lc];
	yin[lc]  = tmp2[iv][lc];
      }
      for (lc=0; lc<=nlyr_common; lc++) 
	xout[lc] = zd_common[lc];
      
      status = regrid (xin, yin, nlyr+1,
		       xout, yout, nlyr_common+1, scale);
      if ( status != 0 ) {
	fprintf(stderr, "Redistribute: error calling regrid for crs.crs_amf\n");
	return status;
      }
      for (lc=0; lc<=nlyr_common; lc++)
	crs->crs_amf[iv][lc] = yout[lc];      
    }
    ASCII_free_double (tmp2, nlambda);

    /** nsza_denstab **/

    /* Allocate memory for temporary ary */
    if ((status = ASCII_calloc_double (&tmp2, 
      				       microphys->nsza_denstab, 
      				       nlyr+1)) != 0)
      return status;

    /* cp original stuff to tmp ary */
    for (lc=0; lc<=nlyr; lc++) 
      for (is=0; is<microphys->nsza_denstab; is++) 
    	tmp2[is][lc] = microphys->denstab_amf[is][lc];

    /* free original ary and allocate with changed resolution */
    ASCII_free_float (microphys->denstab_amf, microphys->nsza_denstab);
    if ((status = ASCII_calloc_float (&microphys->denstab_amf, 
				      microphys->nsza_denstab, 
				      nlyr_common+1)) != 0)
      return status;
    
    for (lc=0; lc<=nlyr; lc++) 
      xfin[nlyr-lc] = zd[lc];
    for (lc=0; lc<=nlyr_common; lc++) 
      xfout[lc]  = zd_common[lc];

    for (is=0; is<microphys->nsza_denstab; is++) {
      for (lc=0; lc<=nlyr; lc++) 
      	yfin[nlyr-lc] = (float) tmp2[is][lc];

      status = arb_wvn (nlyr+1, xfin, yfin,
			nlyr_common+1, xfout, yfout, 1, 0); 
      if (status!=0) {
	fprintf (stderr, "Redistribute: Error %d interpolating denstab_amf\n", status);
	return status;
      }
      for (lc=0; lc<=nlyr_common; lc++) 
	microphys->denstab_amf[is][lc] = yfout[lc];
    }
    
    ASCII_free_double (tmp2, microphys->nsza_denstab);

  }

  /* free memory */
  free (xin); free (xout);
  free (yin); free (yout);

  free (xfin); free (xfout);
  free (yfin); free (yfout);



  return 0;
}
